package dev.fyloz.colorrecipesexplorer.service

import com.nhaarman.mockitokotlin2.*
import dev.fyloz.colorrecipesexplorer.exception.AlreadyExistsException
import dev.fyloz.colorrecipesexplorer.model.*
import dev.fyloz.colorrecipesexplorer.model.account.group
import dev.fyloz.colorrecipesexplorer.repository.RecipeRepository
import dev.fyloz.colorrecipesexplorer.service.files.FileService
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.TestInstance
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import org.springframework.web.multipart.MultipartFile
import java.io.File
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

@TestInstance(TestInstance.Lifecycle.PER_CLASS)
class RecipeServiceTest :
    AbstractExternalModelServiceTest<Recipe, RecipeSaveDto, RecipeUpdateDto, RecipeService, RecipeRepository>() {
    override val repository: RecipeRepository = mock()
    private val companyService: CompanyService = mock()
    private val mixService: MixService = mock()
    private val groupService: GroupService = mock()
    private val recipeStepService: RecipeStepService = mock()
    override val service: RecipeService =
        spy(RecipeServiceImpl(repository, companyService, mixService, recipeStepService, groupService, mock(), mock()))

    private val company: Company = company(id = 0L)
    override val entity: Recipe = recipe(id = 0L, name = "recipe", company = company)
    override val anotherEntity: Recipe = recipe(id = 1L, name = "another recipe", company = company)
    override val entitySaveDto: RecipeSaveDto = spy(recipeSaveDto(name = entity.name, companyId = entity.company.id!!))
    override val entityUpdateDto: RecipeUpdateDto = spy(recipeUpdateDto(id = entity.id!!, name = entity.name))

    @AfterEach
    override fun afterEach() {
        reset(companyService, mixService)
        super.afterEach()
    }

    // existsByCompany()

    @Test
    fun `existsByCompany() returns true when at least one recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(true)

        val found = service.existsByCompany(company)

        assertTrue(found)
    }

    @Test
    fun `existsByCompany() returns false when no recipe exists for the given company`() {
        whenever(repository.existsByCompany(company)).doReturn(false)

        val found = service.existsByCompany(company)

        assertFalse(found)
    }

    // existsByNameAndCompany()

    @Test
    fun `existsByNameAndCompany() returns if a recipe exists for the given name and company in the repository`() {
        setOf(true, false).forEach {
            whenever(repository.existsByNameAndCompany(entity.name, company)).doReturn(it)

            val exists = service.existsByNameAndCompany(entity.name, company)

            assertEquals(it, exists)
        }
    }

    // getAllByName()

    @Test
    fun `getAllByName() returns the recipes with the given name`() {
        val recipes = listOf(entity, anotherEntity)

        whenever(repository.findAllByName(entity.name)).doReturn(recipes)

        val found = service.getAllByName(entity.name)

        assertEquals(recipes, found)
    }

    // getAllByCompany()

    @Test
    fun `getAllByCompany() returns the recipes with the given company`() {
        val companies = listOf(entity, anotherEntity)
        whenever(repository.findAllByCompany(company)).doReturn(companies)

        val found = service.getAllByCompany(company)

        assertEquals(companies, found)
    }

    // save()

    @Test
    override fun `save(dto) calls and returns save() with the created entity`() {
        whenever(companyService.getById(company.id!!)).doReturn(company)
        doReturn(false).whenever(service).existsByNameAndCompany(entity.name, company)
        withBaseSaveDtoTest(entity, entitySaveDto, service, { argThat { this.id == null && this.color == color } })
    }

    @Test
    fun `save(dto) throw AlreadyExistsException when a recipe with the given name and company exists in the repository`() {
        whenever(companyService.getById(company.id!!)).doReturn(company)
        doReturn(true).whenever(service).existsByNameAndCompany(entity.name, company)

        with(assertThrows<AlreadyExistsException> { service.save(entitySaveDto) }) {
            this.assertErrorCode("company-name")
        }
    }

    // update()

    @Test
    override fun `update(dto) calls and returns update() with the created entity`() {
        doReturn(false).whenever(service).existsByNameAndCompany(entity.name, company)
        withBaseUpdateDtoTest(entity, entityUpdateDto, service, { any() })
    }

    @Test
    fun `update(dto) throws AlreadyExistsException when a recipe exists for the given name and company`() {
        val name = "another recipe"

        doReturn(entity).whenever(service).getById(entity.id!!)
        doReturn(true).whenever(service).existsByNameAndCompany(name, company)
        doReturn(name).whenever(entityUpdateDto).name

        with(assertThrows<AlreadyExistsException> { service.update(entityUpdateDto) }) {
            this.assertErrorCode("company-name")
        }
    }

    // updatePublicData()

    @Test
    fun `updatePublicData() updates the notes of a recipe groups information according to the RecipePublicDataDto`() {
        val recipe = recipe(
            id = 0L, groupsInformation = setOf(
                recipeGroupInformation(id = 0L, group = group(id = 1L), note = "Old note"),
                recipeGroupInformation(id = 1L, group = group(id = 2L), note = "Another note"),
                recipeGroupInformation(id = 2L, group = group(id = 3L), note = "Up to date note")
            )
        )
        val notes = setOf(
            noteDto(groupId = 1, content = "Note 1"),
            noteDto(groupId = 2, content = null)
        )
        val publicData = recipePublicDataDto(recipeId = recipe.id!!, notes = notes)

        doReturn(recipe).whenever(service).getById(recipe.id!!)
        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        service.updatePublicData(publicData)

        verify(service).update(argThat<Recipe> {
            assertTrue { this.groupsInformation.first { it.group.id == 1L }.note == notes.first { it.groupId == 1L }.content }
            assertTrue { this.groupsInformation.first { it.group.id == 2L }.note == notes.first { it.groupId == 2L }.content }
            assertTrue { this.groupsInformation.any { it.group.id == 3L } && this.groupsInformation.first { it.group.id == 3L }.note == null }
            true
        })
        verify(mixService, times(0)).updateLocations(any())
    }

    @Test
    fun `updatePublicData() update the location of a recipe mixes in the mix service according to the RecipePublicDataDto`() {
        val publicData = recipePublicDataDto(
            mixesLocation = setOf(
                mixLocationDto(mixId = 0L, location = "Loc 1"),
                mixLocationDto(mixId = 1L, location = "Loc 2")
            )
        )

        service.updatePublicData(publicData)

        verify(mixService).updateLocations(publicData.mixesLocation!!)
        verify(service, times(0)).update(any<Recipe>())
    }

    // addMix()

    @Test
    fun `addMix() adds the given mix to the given recipe and updates it`() {
        val mix = mix(id = 0L)
        val recipe = recipe(id = 0L, mixes = mutableListOf())

        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        val found = service.addMix(recipe, mix)

        verify(service).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertTrue(found.mixes.contains(mix))
    }

    // removeMix()

    @Test
    fun `removeMix() removes the given mix from its recipe and updates it`() {
        val recipe = recipe(id = 0L, mixes = mutableListOf())
        val mix = mix(id = 0L, recipe = recipe)
        recipe.mixes.add(mix)

        doAnswer { it.arguments[0] }.whenever(service).update(any<Recipe>())

        val found = service.removeMix(mix)

        verify(service).update(any<Recipe>())

        assertEquals(recipe.id, found.id)
        assertFalse(found.mixes.contains(mix))
    }
}

private class RecipeImageServiceTestContext {
    val fileService = mockk<FileService> {
        every { write(any<MultipartFile>(), any(), any()) } just Runs
        every { delete(any()) } just Runs
    }
    val recipeImageService = spyk(RecipeImageServiceImpl(fileService))
    val recipe = spyk(recipe())
    val recipeImagesIds = setOf(1L, 10L, 21L)
    val recipeImagesNames = recipeImagesIds.map { it.imageName }.toSet()
    val recipeImagesFiles = recipeImagesNames.map { File(it) }.toTypedArray()
    val recipeDirectory = spyk(File(recipe.imagesDirectoryPath)) {
        every { exists() } returns true
        every { isDirectory } returns true
        every { listFiles() } returns recipeImagesFiles
    }

    init {
        with(recipeImageService) {
            every { recipe.getDirectory() } returns recipeDirectory
        }
    }

    val Long.imageName
        get() = "${recipe.name}$RECIPE_IMAGE_ID_DELIMITER$this"

    val String.imagePath
        get() = "${recipe.imagesDirectoryPath}/$this$RECIPE_IMAGE_EXTENSION"
}

class RecipeImageServiceTest {
    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun test(test: RecipeImageServiceTestContext.() -> Unit) {
        RecipeImageServiceTestContext().test()
    }

    // getAllImages()

    @Test
    fun `getAllImages() returns a Set containing the name of every files in the recipe's directory`() {
        test {
            val foundImagesNames = recipeImageService.getAllImages(recipe)

            assertEquals(recipeImagesNames, foundImagesNames)
        }
    }

    @Test
    fun `getAllImages() returns an empty Set when the recipe's directory does not exists`() {
        test {
            every { recipeDirectory.exists() } returns false

            assertTrue {
                recipeImageService.getAllImages(recipe).isEmpty()
            }
        }
    }

    // download()

    @Test
    fun `download() writes the given image to the FileService and returns its name`() {
        test {
            val mockImage = MockMultipartFile("image.jpg", byteArrayOf(*"Random data".encodeToByteArray()))
            val expectedImageId = recipeImagesIds.maxOrNull()!! + 1L
            val expectedImageName = expectedImageId.imageName
            val expectedImagePath = expectedImageName.imagePath

            val foundImageName = recipeImageService.download(mockImage, recipe)

            assertEquals(expectedImageName, foundImageName)

            verify {
                fileService.write(mockImage, expectedImagePath, true)
            }
        }
    }

    // delete()

    @Test
    fun `delete() deletes the image with the given name in the FileService`() {
        test {
            val imageName = recipeImagesIds.first().imageName
            val imagePath = imageName.imagePath

            recipeImageService.delete(recipe, imageName)

            verify {
                fileService.delete(imagePath)
            }
        }
    }
}
